"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var index_1 = require("typedoc/dist/lib/models/index");
var enumerations_1 = require("../enumerations");
var typedoc_utils_1 = require("../typedoc/typedoc_utils");
/**
 * Class that generates PlantUML code.
 */
var PlantUmlCodeGenerator = /** @class */ (function () {
    /**
     * Creates a new PlantUmlGenerator object with the given options.
     * @param options The options for the PlantUML output.
     */
    function PlantUmlCodeGenerator(options) {
        this.options = options;
    }
    /**
     * Generates the Plant UML lines for the class diagram of the given reflection.
     * @param reflection The reflection for which to generate a class diagram.
     * @returns The Plant UML lines for the class diagram of the given reflection.
     *          If the given reflection is not part of an inheritance or implementation, the result is an empty array.
     */
    PlantUmlCodeGenerator.prototype.createClassDiagramPlantUmlForReflection = function (reflection) {
        var includeMembers = this.options.classDiagramType === enumerations_1.ClassDiagramType.Detailed;
        var plantUmlLines = new Array();
        var siblingsAbove = 0;
        var siblingsBelow = 0;
        // add class/interface
        plantUmlLines = plantUmlLines.concat(this.createPlantUmlForReflection(reflection, includeMembers));
        // add classes/interfaces this type is extending
        var extendedTypes = typedoc_utils_1.TypeDocUtils.getExtendedTypesForReflection(reflection);
        for (var _i = 0, extendedTypes_1 = extendedTypes; _i < extendedTypes_1.length; _i++) {
            var extendedType = extendedTypes_1[_i];
            plantUmlLines = plantUmlLines.concat(this.createPlantUmlForType(extendedType, includeMembers));
            plantUmlLines.push(this.escapeName(extendedType.name) + " <|-- " + this.escapeName(reflection.name));
            ++siblingsAbove;
        }
        // add classes/interfaces this type is implementing
        var implementedTypes = typedoc_utils_1.TypeDocUtils.getImplementedTypesForReflection(reflection);
        for (var _a = 0, implementedTypes_1 = implementedTypes; _a < implementedTypes_1.length; _a++) {
            var implementedType = implementedTypes_1[_a];
            plantUmlLines = plantUmlLines.concat(this.createPlantUmlForType(implementedType, includeMembers));
            plantUmlLines.push(this.escapeName(implementedType.name) + " <|.. " + this.escapeName(reflection.name));
            ++siblingsAbove;
        }
        // add classes/interfaces that are extending this type
        var extendedBys = typedoc_utils_1.TypeDocUtils.getExtendedBysForReflection(reflection);
        for (var _b = 0, extendedBys_1 = extendedBys; _b < extendedBys_1.length; _b++) {
            var extendedBy = extendedBys_1[_b];
            plantUmlLines = plantUmlLines.concat(this.createPlantUmlForType(extendedBy, includeMembers));
            plantUmlLines.push(this.escapeName(reflection.name) + " <|-- " + this.escapeName(extendedBy.name));
            ++siblingsBelow;
        }
        // add classes that are implementing this type
        var implementedBys = typedoc_utils_1.TypeDocUtils.getImplementedBysForReflection(reflection);
        for (var _c = 0, implementedBys_1 = implementedBys; _c < implementedBys_1.length; _c++) {
            var implementedBy = implementedBys_1[_c];
            plantUmlLines = plantUmlLines.concat(this.createPlantUmlForType(implementedBy, includeMembers));
            plantUmlLines.push(this.escapeName(reflection.name) + " <|.. " + this.escapeName(implementedBy.name));
            ++siblingsBelow;
        }
        if (this.options.classDiagramHideEmptyMembers) {
            plantUmlLines.unshift("hide empty fields");
            plantUmlLines.unshift("hide empty methods");
        }
        if (this.options.classDiagramHideCircledChar) {
            plantUmlLines.unshift("hide circle");
        }
        if (siblingsAbove > this.options.classDiagramTopDownLayoutMaxSiblings ||
            siblingsBelow > this.options.classDiagramTopDownLayoutMaxSiblings) {
            plantUmlLines.unshift("left to right direction");
        }
        if (this.options.classDiagramMemberVisibilityStyle === enumerations_1.ClassDiagramMemberVisibilityStyle.Text) {
            plantUmlLines.unshift("skinparam ClassAttributeIconSize 0");
        }
        if (this.options.classDiagramHideShadow) {
            plantUmlLines.unshift("skinparam Shadowing false");
        }
        if (this.options.classDiagramBoxBorderRadius) {
            plantUmlLines.unshift("skinparam RoundCorner " + this.options.classDiagramBoxBorderRadius);
        }
        if (this.options.classDiagramBoxBackgroundColor) {
            plantUmlLines.unshift("skinparam ClassBackgroundColor " + this.options.classDiagramBoxBackgroundColor);
        }
        if (this.options.classDiagramBoxBorderColor) {
            plantUmlLines.unshift("skinparam ClassBorderColor " + this.options.classDiagramBoxBorderColor);
        }
        if (this.options.classDiagramBoxBorderWidth >= 0) {
            plantUmlLines.unshift("skinparam ClassBorderThickness " + this.options.classDiagramBoxBorderWidth);
        }
        if (this.options.classDiagramArrowColor) {
            plantUmlLines.unshift("skinparam ClassArrowColor " + this.options.classDiagramArrowColor);
        }
        if (this.options.classDiagramClassFontName) {
            plantUmlLines.unshift("skinparam ClassFontName " + this.options.classDiagramClassFontName);
        }
        if (this.options.classDiagramClassFontSize) {
            plantUmlLines.unshift("skinparam ClassFontSize " + this.options.classDiagramClassFontSize);
        }
        if (this.options.classDiagramClassFontStyle !== "" /* Undefined */) {
            plantUmlLines.unshift("skinparam ClassFontStyle " + this.options.classDiagramClassFontStyle.toString());
        }
        if (this.options.classDiagramClassFontColor) {
            plantUmlLines.unshift("skinparam ClassFontColor " + this.options.classDiagramClassFontColor);
        }
        if (this.options.classDiagramClassAttributeFontName) {
            plantUmlLines.unshift("skinparam ClassAttributeFontName " + this.options.classDiagramClassAttributeFontName);
        }
        if (this.options.classDiagramClassAttributeFontSize) {
            plantUmlLines.unshift("skinparam ClassAttributeFontSize " + this.options.classDiagramClassAttributeFontSize);
        }
        if (this.options.classDiagramClassAttributeFontStyle !== "" /* Undefined */) {
            plantUmlLines.unshift("skinparam ClassAttributeFontStyle " + this.options.classDiagramClassAttributeFontStyle.toString());
        }
        if (this.options.classDiagramClassAttributeFontColor) {
            plantUmlLines.unshift("skinparam ClassAttributeFontColor " + this.options.classDiagramClassAttributeFontColor);
        }
        return plantUmlLines;
    };
    /**
     * Creates an array of PlantUML lines for generating the box (including its properties and methods) of a reflection.
     * @param reflection The reflection for which the PlantUML should be generated.
     * @param includeMembers Specifies whether the resulting PlantUML should include the properties and methods of
     *                       the reflection as well.
     * @returns The PlantUML lines for the reflection.
     */
    PlantUmlCodeGenerator.prototype.createPlantUmlForReflection = function (reflection, includeMembers) {
        var plantUmlLines = new Array();
        if (reflection.kind === index_1.ReflectionKind.Class || reflection.kind === index_1.ReflectionKind.Interface) {
            plantUmlLines.push(this.createPlantUmlForClassOrInterface(reflection) + " {");
            if (includeMembers && reflection.children) {
                for (var _i = 0, _a = reflection.children; _i < _a.length; _i++) {
                    var children = _a[_i];
                    if (children.kind === index_1.ReflectionKind.Property) {
                        plantUmlLines.push(this.createPlantUmlForProperty(children));
                    }
                    else if (children.kind === index_1.ReflectionKind.Method && children.signatures) {
                        for (var _b = 0, _c = children.signatures; _b < _c.length; _b++) {
                            var signature = _c[_b];
                            plantUmlLines.push(this.createPlantUmlForMethod(signature));
                        }
                    }
                }
            }
            plantUmlLines.push("}");
        }
        return plantUmlLines;
    };
    /**
     * Creates an array of PlantUML lines for generating the box (including its properties and methods) of a type.
     * @param type The type for which the PlantUML should be generated.
     * @param includeMembers Specifies whether the resulting PlantUML should include the properties and methods of
     *                       the type as well.
     * @returns The PlantUML lines for the type.
     */
    PlantUmlCodeGenerator.prototype.createPlantUmlForType = function (type, includeMembers) {
        var reflection = type.reflection;
        if (!reflection) {
            // create a dummy definition for the type which has no reflection
            var name = this.escapeName(type.toString());
            var code = new Array();
            code.push("hide " + name + " circle"); // hide the circle, because we don't know if it is really a class
            code.push("class " + name + " {");
            // Older PlantUML versions raise a syntax error when the class body is empty. So we add a comment:
            code.push("    ' reflection not available");
            code.push("}");
            return code;
        }
        else if (reflection instanceof index_1.DeclarationReflection) {
            return this.createPlantUmlForReflection(reflection, includeMembers);
        }
        return [];
    };
    /**
     * Returns the PlantUML line for generating the output for a given property.
     * @param property The property for which the PlantUML should be generated.
     * @returns The PlantUML line for the given property.
     */
    PlantUmlCodeGenerator.prototype.createPlantUmlForProperty = function (property) {
        var plantUml = "    "; // indent
        if (property.flags.isStatic) {
            plantUml += "{static} ";
        }
        if (property.flags.isPrivate) {
            plantUml += "-";
        }
        else if (property.flags.isProtected) {
            plantUml += "#";
        }
        else {
            plantUml += "+"; // default is public for JS/TS
        }
        plantUml += property.name + " : " + (property.type ? property.type.toString() : "unknown");
        return plantUml;
    };
    /**
     * Returns the PlantUML line for generating the output for a given method.
     * @param methode The method for which the PlantUML should be generated.
     * @returns The PlantUML line for the given method.
     */
    PlantUmlCodeGenerator.prototype.createPlantUmlForMethod = function (method) {
        var plantUml = "    "; // indent
        if (method.flags.isStatic) {
            plantUml += "{static} ";
        }
        if (method.flags.isAbstract) {
            plantUml += "{abstract} ";
        }
        if (method.flags.isPrivate) {
            plantUml += "-";
        }
        else if (method.flags.isProtected) {
            plantUml += "#";
        }
        else {
            plantUml += "+"; // public is default for JS/TS
        }
        plantUml += method.name + "(";
        if (method.parameters) {
            if (this.options.classDiagramMethodParameterOutput === enumerations_1.MethodParameterOutput.OnlyNames) {
                plantUml += method.parameters.map(function (p) { return p.name; }).join(", ");
            }
            else if (this.options.classDiagramMethodParameterOutput === enumerations_1.MethodParameterOutput.OnlyTypes) {
                plantUml += method.parameters.map(function (p) { return (p.type ? p.type.toString() : "unknown"); }).join(", ");
            }
            else if (this.options.classDiagramMethodParameterOutput === enumerations_1.MethodParameterOutput.Complete) {
                plantUml += method.parameters
                    .map(function (p) { return p.name + ": " + (p.type ? p.type.toString() : "unknown"); })
                    .join(", ");
            }
        }
        plantUml += ")";
        if (method.type) {
            plantUml += " : " + method.type.toString();
        }
        else {
            plantUml += " : void";
        }
        return plantUml;
    };
    /**
     * Returns the PlantUML line for the introduction of a class or interface.
     * @param reflection The class or interface for which the PlantUML should be generated.
     * @returns The PlantUML line for the given class or interface.
     */
    PlantUmlCodeGenerator.prototype.createPlantUmlForClassOrInterface = function (reflection) {
        var plantUml = "";
        if (reflection.flags.isStatic) {
            plantUml += "static ";
        }
        if (reflection.flags.isAbstract) {
            plantUml += "abstract ";
        }
        if (reflection.kind === index_1.ReflectionKind.Class) {
            plantUml += "class ";
        }
        else {
            plantUml += "interface ";
        }
        plantUml += this.escapeName(reflection.name);
        return plantUml;
    };
    /**
     * Escapes the name by putting double quotes around it preventing a PlantUML syntax error.
     * @param name The original name.
     * @returns The escaped name.
     */
    PlantUmlCodeGenerator.prototype.escapeName = function (name) {
        return '"' + name + '"';
    };
    return PlantUmlCodeGenerator;
}());
exports.PlantUmlCodeGenerator = PlantUmlCodeGenerator;
