"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var fs = require("fs");
var path = require("path");
var ProgressBar = require("progress");
var typedoc_1 = require("typedoc");
var converter_1 = require("typedoc/dist/lib/converter");
var events_1 = require("typedoc/dist/lib/output/events");
var enumerations_1 = require("./enumerations");
var image_url_generator_1 = require("./image_url_generator");
var logger_1 = require("./logger");
var caching_plantuml_code_generator_1 = require("./plantuml/caching_plantuml_code_generator");
var plantuml_diagram_generator_1 = require("./plantuml/plantuml_diagram_generator");
var plugin_options_1 = require("./plugin_options");
var page_processor_1 = require("./typedoc/page_processor");
var page_section_1 = require("./typedoc/page_section");
var page_section_finder_1 = require("./typedoc/page_section_finder");
var typedoc_utils_1 = require("./typedoc/typedoc_utils");
/**
 * The UML class diagram generator plugin.
 *
 * # What does it do?
 *
 * This plugin generates an UML class diagram for classes and interfaces.
 *
 * # How does it do it?
 *
 * 1. The plugin scans through the classes and interfaces and generates PlantUML code for them.
 * 2. The plugin uses the node-plantuml module to generate image files from the PlantUML code.
 * 3. The plugin adds a section with the image to the page of each class and interface.
 */
var Plugin = /** @class */ (function () {
    function Plugin() {
        /** The options of this plugin. */
        this.options = new plugin_options_1.PluginOptions();
        /** Object that uses PlantUML code to generate PlantUML diagrams. */
        this.plantUmlDiagramGenerator = new plantuml_diagram_generator_1.PlantUmlDiagramGenerator();
        /** Objet that generates URLs for PlantUML diagrams.  */
        this.imageUrlGenerator = new image_url_generator_1.ImageUrlGenerator();
    }
    Object.defineProperty(Plugin.prototype, "isActive", {
        /**
         * Checks if the plugin is active and should generate output.
         * @returns True if the plugin is active, otherwise false.
         */
        get: function () {
            return (this.options.classDiagramType === enumerations_1.ClassDiagramType.Simple ||
                this.options.classDiagramType === enumerations_1.ClassDiagramType.Detailed);
        },
        enumerable: true,
        configurable: true
    });
    /**
     * Initializes the plugin.
     * @param typedoc The TypeDoc application.
     */
    Plugin.prototype.initialize = function (typedoc) {
        this.addOptionsToApplication(typedoc);
        this.subscribeToApplicationEvents(typedoc);
    };
    /**
     * Adds the plugin's options to the application's options.
     * @param typedoc The TypeDoc application.
     */
    Plugin.prototype.addOptionsToApplication = function (typedoc) {
        this.options.addToApplication(typedoc);
    };
    /**
     * Subscribes to events of the application so that the plugin can do its work
     * in the particular doc generation phases.
     * @param typedoc The TypeDoc application.
     */
    Plugin.prototype.subscribeToApplicationEvents = function (typedoc) {
        var _this = this;
        typedoc.converter.on(converter_1.Converter.EVENT_RESOLVE_BEGIN, function (c) { return _this.onConverterResolveBegin(c); });
        typedoc.converter.on(converter_1.Converter.EVENT_RESOLVE_END, function (c) { return _this.onConverterResolveEnd(c); });
        typedoc.renderer.on(events_1.RendererEvent.BEGIN, function (e) { return _this.onRendererBegin(e); });
        typedoc.renderer.on(events_1.PageEvent.END, function (e) { return _this.onRendererEndPage(e); });
        typedoc.renderer.on(events_1.RendererEvent.END, function (e) { return _this.onRendererEnd(e); });
    };
    /**
     * Triggered when the TypeDoc converter begins resolving a project.
     * Reads plugin parameter values and init members.
     * @param context Describes the current state the converter is in.
     */
    Plugin.prototype.onConverterResolveBegin = function (context) {
        this.options.readValuesFromApplication(context.converter.owner.application);
        if (this.isActive) {
            this.plantUmlCodeGenerator = new caching_plantuml_code_generator_1.CachingPlantUmlCodeGenerator(this.options);
            if (this.options.createVerboseOutput) {
                this.log = new logger_1.Logger();
            }
        }
    };
    /**
     * Triggered when the TypeDoc converter has finished resolving a project.
     * Setup progress bar.
     * @param context Describes the current state the converter is in.
     */
    Plugin.prototype.onConverterResolveEnd = function (context) {
        var _a, _b;
        if (this.isActive && !this.options.hideProgressBar) {
            (_a = this.log) === null || _a === void 0 ? void 0 : _a.info("Caculating number of diagrams to generate ...");
            var numberOfDiagramsToGenerate = this.computeDiagramCount(context.project);
            (_b = this.log) === null || _b === void 0 ? void 0 : _b.info("Setting up progress bar ...");
            this.progressBar = new ProgressBar("Adding " + numberOfDiagramsToGenerate + " class diagrams [:bar] :percent", {
                total: numberOfDiagramsToGenerate,
                width: 40,
            });
        }
    };
    /**
     * Computes how many diagrams the plugin must generate.
     * @param project The project containing the reflections for which the diagrams are generated.
     * @returns The number of diagrams to generate.
     */
    Plugin.prototype.computeDiagramCount = function (project) {
        var count = 0;
        for (var key in project.reflections) {
            var reflection = project.reflections[key];
            if (reflection && this.shouldProcessReflection(reflection)) {
                ++count;
            }
        }
        return count;
    };
    /**
     * Triggered before the renderer starts rendering a project.
     * Setup image generator output directory and progress bar.
     * @param event The event emitted by the renderer class.
     */
    Plugin.prototype.onRendererBegin = function (event) {
        if (this.isActive) {
            this.outputDirectory = path.join(event.outputDirectory, "assets/images/");
        }
    };
    /**
     * Triggered after a document has been rendered, just before it is written to disc.
     * Generates a UML class diagram and adds it to the page of the reflection.
     * @param event The event emitted by the renderer class.
     */
    Plugin.prototype.onRendererEndPage = function (event) {
        var _a, _b, _c, _d;
        if (this.isActive && this.shouldProcessPage(event.contents, event.model)) {
            (_a = this.log) === null || _a === void 0 ? void 0 : _a.info("Processing page for reflection " + ((_b = event.model) === null || _b === void 0 ? void 0 : _b.name) + " ...");
            this.processPage(event);
        }
        else {
            (_c = this.log) === null || _c === void 0 ? void 0 : _c.info("Skipping page for reflection " + ((_d = event.model) === null || _d === void 0 ? void 0 : _d.name) + " ...");
        }
    };
    /**
     * Checks if the plugin should process a given page.
     * @param pageContent The content of the page.
     * @param pageModel The model for the page.
     * @returns True, if the plugin should process the page, otherwise false.
     */
    Plugin.prototype.shouldProcessPage = function (pageContent, pageModel) {
        if (pageContent &&
            this.shouldProcessReflection(pageModel) &&
            page_section_finder_1.PageSectionFinder.hasSection(pageContent, "Hierarchy" /* Hierarchy */)) {
            return true;
        }
        return false;
    };
    /**
     * Checks if the plugin should process a given reflection.
     * @param reflection The reflection in question.
     * @returns True, if the plugin should process the reflection, otherwise false.
     */
    Plugin.prototype.shouldProcessReflection = function (reflection) {
        if (reflection instanceof typedoc_1.DeclarationReflection &&
            (reflection.kind === typedoc_1.ReflectionKind.Class || reflection.kind === typedoc_1.ReflectionKind.Interface) &&
            typedoc_utils_1.TypeDocUtils.reflectionIsPartOfClassHierarchy(reflection)) {
            return true;
        }
        return false;
    };
    /**
     * Processes the page by inserting a hierarchy diagram into it.
     * @param event The page event with the page data.
     */
    Plugin.prototype.processPage = function (event) {
        var _this = this;
        var _a, _b;
        var reflection = event.model;
        (_a = this.log) === null || _a === void 0 ? void 0 : _a.info("Generating PlantUML code for reflection " + reflection.name + " ...");
        var plantUml = this.plantUmlCodeGenerator.createClassDiagramPlantUmlForReflection(reflection).join("\n");
        if (this.options.createPlantUmlFiles) {
            (_b = this.log) === null || _b === void 0 ? void 0 : _b.info("Writing PlantUML file for reflection " + reflection.name + " ...");
            this.writePlantUmlFileForReflection(plantUml, reflection);
        }
        var imageUrlPromise;
        if (this.options.outputImageLocation === plugin_options_1.ImageLocation.Local) {
            imageUrlPromise = this.plantUmlDiagramGenerator
                .generateFromCode(plantUml, this.options.outputImageFormat)
                .then(function (result) {
                var _a, _b;
                (_a = _this.log) === null || _a === void 0 ? void 0 : _a.info("Writing local image file for reflection " + reflection.name + " ...");
                var absoluteImageFilePath = _this.writeLocalImageFileForReflection(result, reflection);
                (_b = _this.log) === null || _b === void 0 ? void 0 : _b.info("Creating local image URL for reflection " + reflection.name + " ...");
                return _this.imageUrlGenerator.createLocalImageFileUrl(event.filename, absoluteImageFilePath);
            });
        }
        else if (this.options.outputImageLocation === plugin_options_1.ImageLocation.Embed) {
            imageUrlPromise = this.plantUmlDiagramGenerator
                .generateFromCode(plantUml, this.options.outputImageFormat)
                .then(function (result) {
                var _a;
                (_a = _this.log) === null || _a === void 0 ? void 0 : _a.info("Creating embedded image URL for reflection " + reflection.name + " ...");
                return _this.imageUrlGenerator.createEmbeddedImageUrl(result, _this.options.outputImageFormat);
            });
        }
        else {
            imageUrlPromise = new Promise(function (resolve) {
                var _a;
                (_a = _this.log) === null || _a === void 0 ? void 0 : _a.info("Creating remote image URL for reflection " + reflection.name + " ...");
                resolve(_this.imageUrlGenerator.createPlantUmlServerUrl(plantUml, _this.options.outputImageFormat));
            });
        }
        imageUrlPromise
            .then(function (imageUrl) {
            var _a, _b;
            (_a = _this.log) === null || _a === void 0 ? void 0 : _a.info("Inserting diagram into page for reflection " + reflection.name + " ...");
            _this.insertHierarchyDiagramIntoFile(event.filename, reflection.name, imageUrl);
            (_b = _this.progressBar) === null || _b === void 0 ? void 0 : _b.tick();
        })
            .catch(function (e) {
            var _a;
            (_a = _this.log) === null || _a === void 0 ? void 0 : _a.error("Failed inserting diagram into page for reflection " + reflection.name + ": " + e.message);
        });
    };
    /**
     * Creates a local file with the class diagram of a reflection.
     * @param imageData The image data of the class diagram.
     * @param reflection The reflection for which the file is written.
     * @returns The absolute path to the file that was created.
     */
    Plugin.prototype.writeLocalImageFileForReflection = function (imageData, reflection) {
        var filename = reflection.name + "-umlClassDiagram-" + reflection.id + "." + this.options.outputImageFormat;
        var absoluteFilePath = path.join(this.outputDirectory, filename);
        fs.writeFileSync(absoluteFilePath, imageData);
        return absoluteFilePath;
    };
    /**
     * Creates a file containing the PlantUML code.
     * @param plantUml The PlantUML code that should be written into the file.
     * @param reflection The reflection for which the file is written.
     */
    Plugin.prototype.writePlantUmlFileForReflection = function (plantUml, reflection) {
        var filename = reflection.name + "-umlClassDiagram-" + reflection.id + ".puml";
        var absoluteFilePath = path.join(this.outputDirectory, filename);
        var fileContent = "@startuml\n" + plantUml + "\n@enduml";
        fs.writeFileSync(absoluteFilePath, fileContent, "utf8");
    };
    /**
     * Inserts the hierarchy diagram into the page file.
     * @param filePath Absolute path to the page file.
     * @param reflectionName The name of the reflection for which the diagram was generated.
     * @param imageUrl The URL to the image of the diagram.
     */
    Plugin.prototype.insertHierarchyDiagramIntoFile = function (filePath, reflectionName, imageUrl) {
        var fileContent = fs.readFileSync(filePath, "utf8");
        fileContent = this.insertHierarchyDiagramIntoContent(fileContent, reflectionName, imageUrl);
        fs.writeFileSync(filePath, fileContent, "utf8");
    };
    /**
     * Inserts the hierarchy diagram into the content string.
     * @param originalContent The content string into which the diagram is inserted.
     * @param reflectionName The name of the reflection for which the diagram was generated.
     * @param imageUrl The URL to the image of the diagram.
     * @returns The updated content string with the diagram inserted into it.
     */
    Plugin.prototype.insertHierarchyDiagramIntoContent = function (originalContent, reflectionName, imageUrl) {
        var page = new page_processor_1.PageProcessor(originalContent);
        var hierarchyDiagramSection = page_section_1.PageSection.createForHierarchyDiagram(this.options.sectionTitle, imageUrl, reflectionName);
        if (this.options.classDiagramPosition === plugin_options_1.ClassDiagramPosition.Above) {
            page.insertAboveSection("Hierarchy" /* Hierarchy */, hierarchyDiagramSection);
        }
        else {
            page.insertBelowSection("Hierarchy" /* Hierarchy */, hierarchyDiagramSection);
        }
        return page.content;
    };
    /**
     * Triggered after the renderer has written all documents.
     * Appends style data to the main CSS file.
     * @param event The event emitted by the renderer class.
     */
    Plugin.prototype.onRendererEnd = function (event) {
        var _a, _b;
        if (this.isActive) {
            (_a = this.log) === null || _a === void 0 ? void 0 : _a.info("Attaching content to main.css file ...");
            var filename = path.join(event.outputDirectory, "assets/css/main.css");
            var data = fs.readFileSync(filename, "utf8") +
                "\n.uml-class { max-width:100%; display:block; margin:0 auto; text-align:center }\n";
            fs.writeFileSync(filename, data, "utf8");
            (_b = this.log) === null || _b === void 0 ? void 0 : _b.info("DONE");
        }
    };
    return Plugin;
}());
exports.Plugin = Plugin;
